unit vector;

{$mode objfpc}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  StdCtrls, Grids, ExtCtrls, ExtDlgs, math
  , OpenGLContext, GL, GLU //3D graphics
  , dynmatrix, dynmatrixutils //Matrix math
;

type

  { TVectorForm }

  TVectorForm = class(TForm)
    SmoothedCheckBox: TCheckBox;
    GetMagCalButton: TButton;
    IdleTimer1: TIdleTimer;
    Image1: TImage;
    MagOpenGLControl: TOpenGLControl;
    OpenGLVersion: TLabeledEdit;
    SrcBlendRadioGroup: TRadioGroup;
    DstBlendRadioGroup: TRadioGroup;
    SetMagCalButton: TButton;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    SampleTempButton: TButton;
    SampleMagButton: TButton;
    SampleAccelButton: TButton;
    AStringGrid: TStringGrid;
    XStringGrid: TStringGrid;
    WStringGrid: TStringGrid;
    YStringGrid: TStringGrid;
    Valtitude: TLabeledEdit;
    Vazimuth: TLabeledEdit;
    Vmgrid: TStringGrid;
    Vmonitor: TToggleBox;
    Vpitch: TLabeledEdit;
    Vroll: TLabeledEdit;
    VSampleGroupBox: TGroupBox;
    SetP5Button: TButton;
    SetP6Button: TButton;
    SetP3Button: TButton;
    SetP4Button: TButton;
    SetP1Button: TButton;
    SetP2Button: TButton;
    procedure DstBlendRadioGroupClick(Sender: TObject);
    procedure GetMagCalButtonClick(Sender: TObject);
    procedure IdleTimer1Timer(Sender: TObject);
    procedure srcblendradioChange(Sender: TObject);
    procedure SampleTempButtonClick(Sender: TObject);
    procedure SampleMagButtonClick(Sender: TObject);
    procedure SampleAccelButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SetMagCalButtonClick(Sender: TObject);
    procedure SrcBlendRadioGroupClick(Sender: TObject);
    procedure TimerVTimer(Sender: TObject);
    procedure VmonitorChange(Sender: TObject);
    procedure SetP5ButtonClick(Sender: TObject);
    procedure SetP6ButtonClick(Sender: TObject);
    procedure SetP3ButtonClick(Sender: TObject);
    procedure SetP4ButtonClick(Sender: TObject);
    procedure SetP1ButtonClick(Sender: TObject);
    procedure SetP2ButtonClick(Sender: TObject);
  private
    { private declarations }
    procedure GetAccel();
    procedure Plots();
    procedure GetAccelCal();
    procedure SetAccelCal(ACCpos:Integer);
    procedure ParseAccelCal(result:string);
    procedure CalcAccelCal();
    procedure GetMag();
    procedure GetMagCal();
    procedure ParseMagCal(result:string);
  public
    { public declarations }
    procedure StopMonitoring();

  end;

var
  VectorForm: TVectorForm;
  AccCalXU, AccCalYU,AccCalZU,AccCalXD,AccCalYD,AccCalZD: Integer; //Accelerometer calibration Up/Down
  Axb,Ayb,Azb: Array[0..10] of Double; //Smoothing buffer for accelerometer readings
  Ax,Ay,Az: Double; //Accelerometer raw (or smoothed) readings from meter
  Ax1, Ay1, Az1: Double;  //Accelerometer normalized readings
  Mxmin: Double = 32767;
  Mymin: Double = 32767;
  Mzmin: Double = 32767;
  Mxmax: Double = -32767;
  Mymax: Double = -32767;
  Mzmax: Double = -32767;
  Mxb,Myb,Mzb: Array[0..10] of Double; //Smoothing buffer for magnetometer readings
  Mx,My,Mz: Double; //Magnetic readings from meter
  Mx1, My1, Mz1: Double;  //Compass readings normalized
  Mx2, My2, Mz2: Double;  //Compass readings corrected
  Pitch, Roll: Double;    //Calculated accelerometer values
  Y, w, X: TDMatrix;
  Ynorm: TDMatrix; //Normalized matrix of accelerometer values (Matrix Y)
  Wraw: TDMatrix; //Raw matrix of accelerometer values (Matrix w)

  M: TDMatrix; //Multiplier matrix of accelerometer calibration values
  B: TDMatrix; //Offset matrix of accelerometer calibration values

  Mnorm: TDMatrix; //Normalized matrix of magnetometer values
  Mtilt: TDMatrix; //Tilt coprrected matrix of magnetometer values

  Heading: Double;//Magnetic heading

  srcblend, dstblend: GLenum;

implementation
uses
  Unit1,header_utils,
  uglyfont;

var
//Vector model variables
cube_rotationx: GLFloat;
cube_rotationy: GLFloat;
cube_rotationz: GLFloat;
MagOpenGLControl: TOpenGLControl;

const
ACMmax = 32767;   //Accelerator Compass Module maximum value


{ TVectorForm }

procedure TVectorForm.TimerVTimer(Sender: TObject);
begin
end;

procedure TVectorForm.VmonitorChange(Sender: TObject);
begin
  IdleTimer1.Enabled:=Vmonitor.Checked;
  VSampleGroupBox.Enabled:=not Vmonitor.Checked;
  if Vmonitor.Checked then
    GetAccelCal();
end;

procedure TVectorForm.SetP1ButtonClick(Sender: TObject);
begin
  SetAccelCal(1);
end;
procedure TVectorForm.SetP2ButtonClick(Sender: TObject);
begin
  SetAccelCal(2);
end;

procedure TVectorForm.SetP3ButtonClick(Sender: TObject);
begin
  SetAccelCal(3);
end;
procedure TVectorForm.SetP4ButtonClick(Sender: TObject);
begin
  SetAccelCal(4);
end;
procedure TVectorForm.SetP5ButtonClick(Sender: TObject);
begin
  SetAccelCal(5);
end;
procedure TVectorForm.SetP6ButtonClick(Sender: TObject);
begin
  SetAccelCal(6);
end;

//Initialize this page
procedure TVectorForm.FormCreate(Sender: TObject);
begin

   //Create matrices
   Y:=Mzeros(6,3);
   w:=Mzeros(6,4);
   X:=Mzeros(4,3);
   Ynorm:=Mzeros(3,1);
   Wraw:=Mzeros(3,1);
   M:=Mzeros(3,3);
   B:=Mzeros(3,1);
   Mnorm:=Mzeros(3,1);
   Mtilt:=Mzeros(3,1);

   //Label Y matrix
   YStringGrid.Cells[0,0]:='Y';
   YStringGrid.Cells[1,0]:='x';
   YStringGrid.Cells[2,0]:='y';
   YStringGrid.Cells[3,0]:='z';
   YStringGrid.Cells[0,1]:='P1';
   YStringGrid.Cells[0,2]:='P2';
   YStringGrid.Cells[0,3]:='P3';
   YStringGrid.Cells[0,4]:='P4';
   YStringGrid.Cells[0,5]:='P5';
   YStringGrid.Cells[0,6]:='P6';
   //Fill Y Matrix
   Y:=StringToDMatrix('0.0 0.0 1.0; 0.0 0.0 -1.0; 0.0 1.0 0.0; 0.0 -1.0 0.0; 1.0 0.0 0.0; -1.0 0.0 0.0');
   DMatrixToGrid(YStringGrid, Y);

   //Label W Matrix
   WStringGrid.Cells[0,0]:='W';
   WStringGrid.Cells[1,0]:='x';
   WStringGrid.Cells[2,0]:='y';
   WStringGrid.Cells[3,0]:='z';
   WStringGrid.Cells[0,1]:='P1';
   WStringGrid.Cells[0,2]:='P2';
   WStringGrid.Cells[0,3]:='P3';
   WStringGrid.Cells[0,4]:='P4';
   WStringGrid.Cells[0,5]:='P5';
   WStringGrid.Cells[0,6]:='P6';

   //label Accelerometer readings
   AStringGrid.Cells[0,0]:='Accel';
   AStringGrid.Cells[1,0]:='X';
   AStringGrid.Cells[2,0]:='Y';
   AStringGrid.Cells[3,0]:='Z';
   AStringGrid.Cells[0,1]:='raw';
   AStringGrid.Cells[0,2]:='norm';

   //Label X Matrix
   XStringGrid.Cells[0,1]:='A_1';
   XStringGrid.Cells[0,2]:='A_2';
   XStringGrid.Cells[0,3]:='A_3';
   XStringGrid.Cells[0,4]:='A_0';
   XStringGrid.Cells[1,0]:='A1_';
   XStringGrid.Cells[2,0]:='A2_';
   XStringGrid.Cells[3,0]:='A3_';

   //Label magnetic grid
   Vmgrid.Cells[0,0]:='Magnet';
   Vmgrid.Cells[0,1]:='Max';
   Vmgrid.Cells[0,2]:='M_';
   Vmgrid.Cells[0,3]:='Min';
   Vmgrid.Cells[0,4]:='M_1(norm)';
   Vmgrid.Cells[0,5]:='M_2(corr)';
   Vmgrid.Cells[0,6]:='_h';
   Vmgrid.Cells[1,0]:='X axis';
   Vmgrid.Cells[2,0]:='Y axis';
   Vmgrid.Cells[3,0]:='Z axis';

end;

procedure TVectorForm.FormShow(Sender: TObject);
begin
  //todo **** make sure that these do not get activated when unminimized
  GetAccelCal();
  GetMagCal();
  Plots();

  OpenGLVersion.Text:=glGetString(GL_VERSION);

end;

procedure TVectorForm.SampleTempButtonClick(Sender: TObject);
begin
  SendGet('v0x');
end;

procedure TVectorForm.srcblendradioChange(Sender: TObject);
begin

end;

procedure TVectorForm.DstBlendRadioGroupClick(Sender: TObject);
begin

case DstBlendRadioGroup.ItemIndex of
0: dstblend:=GL_ZERO;
1: dstblend:=GL_ONE;
2: dstblend:=GL_SRC_COLOR;
3: dstblend:=GL_ONE_MINUS_SRC_COLOR;
4: dstblend:=GL_DST_COLOR;
5: dstblend:=GL_ONE_MINUS_DST_COLOR;
6: dstblend:=GL_SRC_ALPHA;
7: dstblend:=GL_ONE_MINUS_SRC_ALPHA;
8: dstblend:=GL_DST_ALPHA;
9: dstblend:=GL_ONE_MINUS_DST_ALPHA;
end;

end;

procedure TVectorForm.GetMagCalButtonClick(Sender: TObject);
begin
  GetMagCal();
end;

procedure TVectorForm.IdleTimer1Timer(Sender: TObject);
begin
  IdleTimer1.Enabled:=False;
  GetAccel(); //gather and compensate accelerometer readings
  GetMag();  //gather and compensate magnetometer readings (uses compensated accelerometer values)
  Plots();
  if Vmonitor.Checked then
    IdleTimer1.Enabled:=True;
end;

procedure TVectorForm.SampleMagButtonClick(Sender: TObject);
begin
  GetMag();
end;

procedure TVectorForm.SampleAccelButtonClick(Sender: TObject);
begin
  GetAccel();
end;

procedure TVectorForm.ParseAccelCal(result:string);
var
   pieces: TStringList;
   AccCalPos: Integer; //Accelerometer position
begin
  pieces := TStringList.Create;
  pieces.Delimiter := ',';

  pieces.DelimitedText := result;
  if ((pieces.Count>=5) and (pieces.Strings[0]='v3')) then
    begin
         AccCalPos:=StrToInt(pieces.Strings[1]);
         w.setv(AccCalPos-1, 0, StrToFloatDef(pieces.Strings[2],0));  //x
         w.setv(AccCalPos-1, 1, StrToFloatDef(pieces.Strings[3],0));  //y
         w.setv(AccCalPos-1, 2, StrToFloatDef(pieces.Strings[4],0));  //z
         w.setv(AccCalPos-1, 3, 1.0);  //constant

         //fill display
         WStringGrid.Cells[1,AccCalPos]:=format('%6.0f',[w.getv(AccCalPos-1, 0)]);
         WStringGrid.Cells[2,AccCalPos]:=format('%6.0f',[w.getv(AccCalPos-1, 1)]);
         WStringGrid.Cells[3,AccCalPos]:=format('%6.0f',[w.getv(AccCalPos-1, 2)]);
         WStringGrid.Cells[4,AccCalPos]:=format('%6.0f',[w.getv(AccCalPos-1, 3)]);
    end
    else
        StatusMessage('V3x get accelerometer calibration failed. Count='+IntToStr(pieces.Count)+'. Reply= '+pieces.Strings[0]);
end;

procedure TVectorForm.SetMagCalButtonClick(Sender: TObject);
Begin
    ParseMagCal(SendGet('v4,'+
                 format('%6.5d,%6.5d,%6.5d,%6.5d,%6.5d,%6.5d',
                 [round(Mxmin), round(Mymin), round(Mzmin), +
                  round(Mxmax), round(Mymax), round(Mzmax) ]) +
                 'x'));
end;

procedure TVectorForm.SrcBlendRadioGroupClick(Sender: TObject);
begin

case SrcBlendRadioGroup.ItemIndex of
0: srcblend:=GL_ZERO;
1: srcblend:=GL_ONE;
2: srcblend:=GL_SRC_COLOR;
3: srcblend:=GL_ONE_MINUS_SRC_COLOR;
4: srcblend:=GL_DST_COLOR;
5: srcblend:=GL_ONE_MINUS_DST_COLOR;
6: srcblend:=GL_SRC_ALPHA;
7: srcblend:=GL_ONE_MINUS_SRC_ALPHA;
8: srcblend:=GL_DST_ALPHA;
9: srcblend:=GL_ONE_MINUS_DST_ALPHA;
13: srcblend:=GL_SRC_ALPHA_SATURATE;
end;

end;

procedure TVectorForm.GetMagCal();
begin
  ParseMagCal(SendGet('v4x'));
end;

procedure TVectorForm.ParseMagCal(result:string);
var
   pieces: TStringList;
   AccCalPos: Integer; //Accelerometer position
begin
  pieces := TStringList.Create;
  pieces.Delimiter := ',';

  pieces.DelimitedText := result;
  if ((pieces.Count=7) and (pieces.Strings[0]='v4')) then
    begin
         Mxmin:=StrToFloatDef(pieces.Strings[1],0);
         Mymin:=StrToFloatDef(pieces.Strings[2],0);
         Mzmin:=StrToFloatDef(pieces.Strings[3],0);
         Mxmax:=StrToFloatDef(pieces.Strings[4],0);
         Mymax:=StrToFloatDef(pieces.Strings[5],0);
         Mzmax:=StrToFloatDef(pieces.Strings[6],0);

         //fill display
         Vmgrid.Cells[1,1]:=format('%7.0f',[Mxmax]);
         Vmgrid.Cells[2,1]:=format('%7.0f',[Mymax]);
         Vmgrid.Cells[3,1]:=format('%7.0f',[Mzmax]);
         Vmgrid.Cells[1,3]:=format('%7.0f',[Mxmin]);
         Vmgrid.Cells[2,3]:=format('%7.0f',[Mymin]);
         Vmgrid.Cells[3,3]:=format('%7.0f',[Mzmin]);
    end
    else
        StatusMessage('V4x get magnetometer calibration failed. Count='+IntToStr(pieces.Count)+'. Reply= '+pieces.Strings[0]);
end;

procedure TVectorForm.GetAccelCal();
var
   AccCalPos: Integer; //Accelerometer position
begin
  for AccCalPos := 1 to 6 do
     ParseAccelCal(SendGet(format('v3,%1dx',[AccCalPos])));

  CalcAccelCal();
end;
procedure TVectorForm.SetAccelCal(ACCpos:Integer);
Begin
  ParseAccelCal(SendGet('v3,'+
               format('%1d,%6.5d,%6.5d,%6.5d',
               [ACCpos,round(Ax), round(Ay), round(Az)]) +
               'x'));
  CalcAccelCal();
end;
procedure TVectorForm.GetAccel();
var
  pieces: TStringList;
  row,col:Integer;
  i : integer;
begin

  pieces := TStringList.Create;
  pieces.Delimiter := ',';

  pieces.DelimitedText := SendGet('v2x',False, 3000, True, True); //Do not show or log value reads

  if SmoothedCheckBox.Checked then
    begin
         //shift
         for i:=0 to 9 do
            begin
              Axb[i+1]:=Axb[i];
              Ayb[i+1]:=Ayb[i];
              Azb[i+1]:=Azb[i];
            end;

         //insert latest reading
         Axb[0]:=StrToFloatDef(pieces.Strings[1],0);
         Ayb[0]:=StrToFloatDef(pieces.Strings[2],0);
         Azb[0]:=StrToFloatDef(pieces.Strings[3],0);

         //clear sum
         Ax:=0;
         Ay:=0;
         Az:=0;

         //sum
         for i:=0 to 10 do
            begin
              Ax:=Ax+Axb[i];
              Ay:=Ay+Ayb[i];
              Az:=Az+Azb[i];
            end;
         //average
         Ax:=Ax/11;
         Ay:=Ay/11;
         Az:=Az/11;
    end
  else
  begin
    Ax:=StrToFloatDef(pieces.Strings[1],0);
    Ay:=StrToFloatDef(pieces.Strings[2],0);
    Az:=StrToFloatDef(pieces.Strings[3],0);
  end;

  //display raw values
  AStringGrid.Cells[1,1]:=format('%7.0f',[Ax]);
  AStringGrid.Cells[2,1]:=format('%7.0f',[Ay]);
  AStringGrid.Cells[3,1]:=format('%7.0f',[Az]);

  //Enable Acceleration calibration buttons
  SetP5Button.Enabled:=Ax >  8000;
  SetP6Button.Enabled  :=Ax < -8000;
  SetP3Button.Enabled:=Ay >  8000;
  SetP4Button.Enabled  :=Ay < -8000;
  SetP1Button.Enabled:=Az >  8000;
  SetP2Button.Enabled  :=Az < -8000;

  //get raw accelerometer values in preparation for normalizing
  Wraw.setv(0, 0, Ax);
  Wraw.setv(1, 0, Ay);
  Wraw.setv(2, 0, Az);

  //Set M (multiplier) values from X
  for row:=0 to 2 do
     for col:=0 to 2 do
        M.setv(col,row, X.getv(row,col)); // Also transpose the matrix while reading

  //Set B (offset) values from X
  for row:=0 to 2 do
     B.setv(row,0, X.getv(3,row));

  //Normalize readings
  Ynorm:=M* Wraw + B;

  //Limit within +/- 1.0
  Ynorm.setv(0, 0, EnsureRange(Ynorm.getv(0, 0),-1.0,1.0));
  Ynorm.setv(1, 0, EnsureRange(Ynorm.getv(1, 0),-1.0,1.0));
  Ynorm.setv(2, 0, EnsureRange(Ynorm.getv(2, 0),-1.0,1.0));

  //Pull out normalized value for convenience
  Ax1:=Ynorm.getv(0, 0);
  Ay1:=Ynorm.getv(1, 0);
  Az1:=Ynorm.getv(2, 0);

  //display normalized values
  AStringGrid.Cells[1,2]:=format('%7.3f',[Ax1]);
  AStringGrid.Cells[2,2]:=format('%7.3f',[Ay1]);
  AStringGrid.Cells[3,2]:=format('%7.3f',[Az1]);

  //Determine Pitch per AN3192.pdf, Equation 10
  //Pitch:=arcsin(-1*Ax1);
  Pitch:=arctan2(Ax1,Az1);
  Vpitch.Text:=format('%4.0f',[radtodeg(Pitch)]);

  //Determine Pitch per AN3192.pdf, Equation 10
  Roll:= arcsin(EnsureRange(Ay1/cos(Pitch),-1.0,1.0));
  VRoll.Text:=format('%7.0f',[radtodeg(Roll)]);

  Valtitude.Text:= format('%4.0f',[radtodeg(arcsin(Ax1))]);

end;
procedure TVectorForm.GetMag();
var
   pieces: TStringList;
   Xh, Yh: Double; //Magnetic strengths corrected for accelerometer angle (Eq 2 AN3192.pdf)
   i : integer;

begin
  pieces := TStringList.Create;
  pieces.Delimiter := ',';

  pieces.DelimitedText := SendGet('v1x',False, 3000, True, True); //Do not show or log value reads

  if SmoothedCheckBox.Checked then
  begin
       //shift
       for i:=0 to 9 do
          begin
            Mxb[i+1]:=Mxb[i];
            Myb[i+1]:=Myb[i];
            Mzb[i+1]:=Mzb[i];
          end;

       //insert latest reading
       Mxb[0]:=StrToFloatDef(pieces.Strings[1],0);
       Myb[0]:=StrToFloatDef(pieces.Strings[2],0);
       Mzb[0]:=StrToFloatDef(pieces.Strings[3],0);

       //clear sum
       Mx:=0;
       My:=0;
       Mz:=0;

       //sum
       for i:=0 to 10 do
          begin
            Mx:=Mx+Mxb[i];
            My:=My+Myb[i];
            MZ:=Mz+Mzb[i];
          end;
       //average
       Mx:=Mx/11;
       My:=My/11;
       Mz:=Mz/11;
  end
else
begin
  Mx:=StrToFloatDef(pieces.Strings[1],0);
  My:=StrToFloatDef(pieces.Strings[2],0);
  Mz:=StrToFloatDef(pieces.Strings[3],0);
end;

  //Compute new min/max magnetic values
  if Mx<Mxmin then Mxmin:= Mx;
  if My<Mymin then Mymin:= My;
  if Mz<Mzmin then Mzmin:= Mz;

  if Mx>Mxmax then Mxmax:= Mx;
  if My>Mymax then Mymax:= My;
  if Mz>Mzmax then Mzmax:= Mz;

  //Normalize readings
  Mx1:=EnsureRange(Mx/((Mxmax-MxMin)/2),-1.0,1.0);
  My1:=EnsureRange(My/((Mymax-MyMin)/2),-1.0,1.0);
  Mz1:=EnsureRange(Mz/((Mzmax-MzMin)/2),-1.0,1.0);

  //Compensate magnetometer readings for 3D position
  //Use equation 12 from AN3192.pdf of LSM303DLH (since there is no similar LSM303D document)
  Mx2:= Mx1*cos(Pitch) + Mz1*sin(Pitch);
  My2:= Mx1*sin(Roll)*sin(Pitch) + My1*cos(Roll) - Mz1*sin(Roll)*cos(Pitch);
  Mz2:= -1.0*Mx1*cos(Roll)*sin(Pitch) + My1*sin(Roll) + Mz1*cos(Roll)*cos(Pitch);

  //Fill up string grid with result from meter
  Vmgrid.Cells[1,1]:=format('%7.0f',[Mxmax]);
  Vmgrid.Cells[2,1]:=format('%7.0f',[Mymax]);
  Vmgrid.Cells[3,1]:=format('%7.0f',[Mzmax]);

  Vmgrid.Cells[1,2]:=format('%7.0f',[Mx]);
  Vmgrid.Cells[2,2]:=format('%7.0f',[My]);
  Vmgrid.Cells[3,2]:=format('%7.0f',[Mz]);

  Vmgrid.Cells[1,3]:=format('%7.0f',[Mxmin]);
  Vmgrid.Cells[2,3]:=format('%7.0f',[Mymin]);
  Vmgrid.Cells[3,3]:=format('%7.0f',[Mzmin]);

  Vmgrid.Cells[1,4]:=format('%3.3f',[Mx1]);
  Vmgrid.Cells[2,4]:=format('%3.3f',[My1]);
  Vmgrid.Cells[3,4]:=format('%3.3f',[Mz1]);

  Vmgrid.Cells[1,5]:=format('%3.3f',[Mx2]);
  Vmgrid.Cells[2,5]:=format('%3.3f',[My2]);
  Vmgrid.Cells[3,5]:=format('%3.3f',[Mz2]);

  //Use equation 13 from AN3192.pdf of LSM303DLH (since there is no similar LSM303D document)
  //if ((Mx2>0) and (My2>=0)) then
  //  Heading:=radtodeg(arctan(My2/Mx2))
  //  else
  //    Heading:=0;
  //if (Mx2<0) then Heading:=radtodeg(pi + arctan(My2/Mx2));
  //if ((Mx2>0) and (My2<=0)) then Heading:=radtodeg(2*pi + arctan(My2/Mx2));
  //if ((Mx2=0) and (My2<0)) then Heading:=radtodeg(pi/2); //90 deg
  //if ((Mx2=0) and (My2>0)) then Heading:=radtodeg(3*pi/2); //270 deg

  //(Eq 2 AN3192.pdf)
  Xh:=Mx1*cos(Pitch) + Mz1*sin(Pitch);
  Yh:=Mx1*sin(Roll)*sin(Pitch) + My1*cos(Roll) - Mz1*sin(Roll)*cos(Pitch);
  //Heading:=arctan(Yh/Xh);

  Vmgrid.Cells[1,6]:=format('%3.3f',[Xh]);
  Vmgrid.Cells[2,6]:=format('%3.3f',[Yh]);


  Mnorm.setv(0, 0, EnsureRange(Mx1,-1.0,1.0));
  Mnorm.setv(1, 0, EnsureRange(My1,-1.0,1.0));
  Mnorm.setv(2, 0, EnsureRange(Mz1,-1.0,1.0));

  Mtilt:=Ynorm + Mnorm;
  Heading:= radtodeg(arctan2(Mtilt.getv(1,0),Mtilt.getv(0,0)));
  if Heading<0 then Heading:=Heading+360;

  Vazimuth.Text:=format('%4.0f',[Heading]);

end;

procedure TVectorForm.Plots();
var
  i: Double;

  procedure cube(X,Y,Z: Double);
  begin
    glPushMatrix();

      //Move drawing to be viewed from above and on an angle
      glTranslatef(0,-0.2,-4);
      glRotatef( 30.0, 1, 0, 0); //X rotation
      glRotatef(-30.0, 0, 1, 0); //Y rotation

      //Draw outer loops
      glLineWidth(0.6);
      glColor4f(1, 1, 1,0.5); //white
      glBegin(GL_LINE_LOOP);
              i:=0;
              while (i<2*pi) do
                    begin
                      glVertex3f( sin(i), 0, cos(i));
                      i:= i+0.1;
                    end;
      glEnd();
      glBegin(GL_LINE_LOOP);
              i:=0;
              while (i<2*pi) do
                    begin
                      glVertex3f( sin(i), cos(i), 0);
                      i:= i+0.001;
                    end;
      glEnd();
      glBegin(GL_LINE_LOOP);
              i:=0;
              while (i<2*pi) do
                    begin
                      glVertex3f( 0, sin(i), cos(i));
                      i:= i+0.1;
                    end;
      glEnd();

      //Draw text axis labels
      glPushMatrix();
        glLineWidth(1.5);
        glRotatef(180, 1, 0, 0);
        glColor4f(1,0,0,0.5); //Red
        glTextOut( 1.2,0,0, 0.15, 0.2, 0.2, 1, 'X');
        glColor4f(0,0,1,0.5); //Blue
        glTextOut( 0,0,-1.2, 0.15, 0.2, 0.2, 1, 'Y');
        glColor4f(0,1,0,0.5); //Green
        glTextOut( 0,-1.2,0, 0.15, 0.2, 0.2, 1, 'Z');
      glPopMatrix();

      //Grid axis lines
      glBegin(GL_LINES);
        glColor4f(1,0,0,0.3);//Red
        glVertex3f(0,0,0);
        glVertex3f(1.1,0,0);
        glColor4f(0,1,0,0.3);//Green
        glVertex3f(0,0,0);
        glVertex3f(0,1.1,0);
        glColor4f(0,0,1,0.3);//Blue
        glVertex3f(0,0,0);
        glVertex3f(0,0,1.1);
      glEnd();

      //Value
      glLineWidth(2);
      glBegin(GL_LINES);
        glColor4f(1,0.5,0.5,0.9); //Red
        glVertex3f(0, 0,0);
        glVertex3f(X, 0,0); // Left/right  value
        glColor4f(0.5, 1, 0.5,0.9); //Green
        glVertex3f(0, 0,0);
        glVertex3f( 0,Z,0); // Up/down Axis
        glColor4f(0.5,0.5,1,0.9); //Blue
        glVertex3f(0, 0,0);
        glVertex3f(0,0, Y); // In/out axis
        glColor4f(1, 1, 1,0.9); //White
        glVertex3f(0, 0,0);
        glVertex3f(X, Z, Y); //Combined vector
      glEnd();

      glPointSize(4);
      glBegin(GL_POINTS);
      glColor4f(1,0.5,0.5,0.9); //Red
      glVertex3f(X, 0,0); // Left/right  value
      glColor4f(0.5, 1, 0.5,0.9); //Green
      glVertex3f( 0,Z,0); // Up/down Axis
      glColor4f(0.5,0.5,1,0.9); //Blue
      glVertex3f(0,0, Y); // In/out axis
      glColor4f(1, 1, 1,0.9); //White
      glVertex3f(X, Z, Y); //Combined vector
      glEnd();

    glPopMatrix();
  end;

  procedure textlabel(text:String);
  begin
    glPushMatrix();
      glLineWidth(1.5);
      //glLineWidth(0.5);
      glColor3f(1, 1, 1); //white
      glTranslatef(0,0.7,-2);
      glRotatef(180, 1, 0, 0);
      glTextOut(0,0,0, 0.07, 0.1, 0.1, 1, text);
    glPopMatrix();
  end;

begin

  //antialiasing
  glEnable (GL_LINE_SMOOTH);
  //glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST); //entire line and mode makes no difference

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  //glBlendFunc(srcblend, dstblend);

  glClearColor(0, 0, 0, 1.0); //Background
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  glEnable(GL_DEPTH_TEST);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, double(100) / 100, 0.1, 100.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  //--- accel plot ------------
  glViewport(0,250,250,250);
  glScissor(0,250,250,250);

  //Draw accelerometer X,Y,Z magnitudes
  textlabel('Accelerometer');
  cube(Ax1,Ay1,Az1);


  //--- mag plot ------------
  glViewport(0,0,250,250);
  glScissor(0,0,250,250);

  //Draw magnetic X,Y,Z magnitudes
  textlabel('Magnetometer');
  cube(Mx1,My1,Mz1);


  //--- AltAz plot ------------
  glViewport(0,500,250,250);
  glScissor(0,500,250,250);

  //Draw final Altitude Azimuth drawing
  //textlabel('Alt Az');
  glPushMatrix();

    //Move drawing to be viewed from above and on an angle
    glTranslatef(0,-0.2,-4);
    glRotatef( 30.0, 1, 0, 0); //X rotation
    glRotatef(-30.0, 0, 1, 0); //Y rotation

    //Draw outer loops
    glLineWidth(0.6);
    glColor4f(1, 1, 1,0.5); //white
    glBegin(GL_LINE_LOOP);
            i:=0;
            while (i<2*pi) do
                  begin
                    glVertex3f( sin(i), 0, cos(i));
                    i:= i+0.1;
                  end;
    glEnd();
    glBegin(GL_LINES);
            i:=0;
            while (i<pi/2) do
                  begin
                    glVertex3f( sin(i), cos(i), 0);
                    i:= i+0.001;
                  end;
    glEnd();

    //Draw text axis labels
    glPushMatrix();
      glLineWidth(1.0);
      glRotatef(180, 1, 0, 0);
      glColor4f(1,1,1,0.5);
      glTextOut( 0,-1.2,0, 0.075, 0.1, 0.1, 1, 'Zenith');
    glPopMatrix();

    //Grid axis lines
    glBegin(GL_LINES);
      glColor4f(1,0,0,0.3);//Red
      glVertex3f(0,0,0);
      glVertex3f(1.1,0,0);
      glColor4f(0,1,0,0.3);//Green
      glVertex3f(0,0,0);
      glVertex3f(0,1.1,0);
    glEnd();

    //Value
    glLineWidth(2);
    glBegin(GL_LINES);
      glColor4f(1,1,0,1); //Yellow
      glVertex3f(0, 0,0);
      //glVertex3f(cos(arcsin(Ax1)),Ax1, 0); // Left/right  value (based on X only) too noisy near X-zero
      glVertex3f(sqrt(Ay1**2+Az1**2),Ax1, 0); // Left/right  value (based on all three axis)
    glEnd();

    //Compass line pointer (to North)
    //glBegin(GL_LINES);
    //  glColor4f(0.5,1.0,0.5,0.9); //Green
    //  glVertex3f(0, 0,0);
    //  glVertex3f(sin(degtorad(Heading)), 0,-1.0*cos(degtorad(Heading))); // Left/right  value
    //glEnd();

    //Compass heading dial (rotating)
    glPushMatrix();
      glLineWidth(1.5);
      glColor4f(0,1,0,1); //Green
      glRotatef(90,1,0,0);
      glRotatef(90,0,0,1);
      glRotatef(-1*Heading, 0,0,1);//Correct direction
      glTextOut( 0, -1.2, 0, 0.15, 0.2, 0.2, 1, 'N');
      glTextOut( 1.2,  0, 0, 0.15, 0.2, 0.2, 1, 'E');
      glTextOut( -1.2, 0, 0, 0.15, 0.2, 0.2, 1, 'W');
      glTextOut( 0,  1.2, 0, 0.15, 0.2, 0.2, 1, 'S');
    glPopMatrix();

  glPopMatrix();


  MagOpenGLControl.SwapBuffers;
end;

procedure TVectorForm.StopMonitoring();
begin
  Vmonitor.Checked:=False;
  IdleTimer1.Enabled:=False;
  VSampleGroupBox.Enabled:=True;
end;

//Calculate accelerometer calibration values
procedure TVectorForm.CalcAccelCal();
begin
  //calibrate accelerometer based on AN3128.pdf for LSM303D
  X:= Minv(w.t * w) * w.t * Y;
  //Fill display
  DMatrixToGrid(XStringGrid, X);
end;


initialization
  {$I vector.lrs}

end.

